/*
 * Copyright (c) 2005-2007 Extreme! Lab, Indiana University. All rights reserved.
 *
 * This software is open source. See the bottom of this file for the license.
 *
 * $Id: Workflow.java,v 1.26 2007/06/12 19:26:53 sshirasu Exp $
 */
package edu.indiana.extreme.xbaya.wf;

import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.net.URI;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import javax.imageio.ImageIO;
import javax.xml.namespace.QName;

import org.apache.commons.codec.binary.Base64;
import org.gpel.model.GpelProcess;
import org.xmlpull.infoset.XmlElement;
import org.xmlpull.infoset.XmlNamespace;

import xsul5.MLogger;
import xsul5.wsdl.WsdlDefinitions;
import edu.indiana.extreme.xbaya.XBayaConstants;
import edu.indiana.extreme.xbaya.XBayaRuntimeException;
import edu.indiana.extreme.xbaya.XBayaVersion;
import edu.indiana.extreme.xbaya.component.Component;
import edu.indiana.extreme.xbaya.component.ComponentException;
import edu.indiana.extreme.xbaya.component.ws.WSComponent;
import edu.indiana.extreme.xbaya.component.ws.WSComponentFactory;
import edu.indiana.extreme.xbaya.component.ws.WSComponentKey;
import edu.indiana.extreme.xbaya.component.ws.WSComponentPort;
import edu.indiana.extreme.xbaya.graph.GraphException;
import edu.indiana.extreme.xbaya.graph.GraphSchema;
import edu.indiana.extreme.xbaya.graph.Node;
import edu.indiana.extreme.xbaya.graph.util.GraphUtil;
import edu.indiana.extreme.xbaya.graph.ws.WSGraph;
import edu.indiana.extreme.xbaya.graph.ws.WSGraphFactory;
import edu.indiana.extreme.xbaya.graph.ws.WSNode;
import edu.indiana.extreme.xbaya.util.StringUtil;
import edu.indiana.extreme.xbaya.util.WSDLUtil;
import edu.indiana.extreme.xbaya.util.XMLUtil;

/**
 * @author Satoshi Shirasuna
 */
public class Workflow {

    /**
     * Namespace prefix
     */
    public static final String NS_PREFIX_XWF = "xwf";

    /**
     * Namespace URI
     */
    public static final String NS_URI_XWF = XBayaConstants.NS_URI_XBAYA + "xwf";

    /**
     * Namespace
     */
    public static final XmlNamespace NS_XWF = XMLUtil.BUILDER.newNamespace(
            NS_PREFIX_XWF, NS_URI_XWF);

    /**
     * WORKFLOW_TAG
     */
    public static final String WORKFLOW_TAG = "workflow";

    private static final String VERSION_ATTRIBUTE = "version";

    private static final String WSDLS_TAG = "wsdls";

    private static final String WSDL_TAG = "wsdl";

    private static final String BPEL_TAG = "bpel";

    private static final String WORKFLOW_WSDL_TAG = "workflowWSDL";

    private static final String IMAGE_TAG = "image";

    private static final String ID_ATTRIBUTE = "id";

    private static final MLogger logger = MLogger.getLogger();

    private WSGraph graph;

    private BufferedImage image;

    private URI gpelTemplateID;

    private URI gpelInstanceID;

    private GpelProcess gpelProcess;

    private WsdlDefinitions workflowWSDL;

    /**
     * used only during the parsing xwf or loading from GPEL.
     */
    private Map<String, WsdlDefinitions> wsdlMap;

    /**
     * Constructs a Workflow.
     */
    public Workflow() {
        this.wsdlMap = new HashMap<String, WsdlDefinitions>();

        // Create a empty graph here to avoid null checks.
        this.graph = WSGraphFactory.createGraph();
    }

    /**
     * Constructs a Workflow.
     * 
     * @param workflowString
     *            Workflow XML in String.
     * @throws GraphException
     * @throws ComponentException
     */
    @Deprecated
    public Workflow(String workflowString) throws GraphException,
            ComponentException {
        this();
        try {
            XmlElement workflowElement = XMLUtil
                    .stringToXmlElement(workflowString);
            parse(workflowElement);
        } catch (RuntimeException e) {
            throw new GraphException(e);
        }
    }

    /**
     * Constructs a Workflow.
     * 
     * @param workflowElement
     * @throws GraphException
     * @throws ComponentException
     */
    public Workflow(XmlElement workflowElement) throws GraphException,
            ComponentException {
        this();
        parse(workflowElement);
    }

    /**
     * @param gpelTemplateID
     */
    public void setGPELTemplateID(URI gpelTemplateID) {
        this.gpelTemplateID = gpelTemplateID;
    }

    /**
     * @return The GPEL Template ID
     */
    public URI getGPELTemplateID() {
        return this.gpelTemplateID;
    }

    /**
     * Returns the gpelInstanceID.
     * 
     * @return The gpelInstanceID
     */
    public URI getGPELInstanceID() {
        return this.gpelInstanceID;
    }

    /**
     * Sets gpelInstanceID.
     * 
     * @param gpelInstanceID
     *            The gpelInstanceID to set.
     */
    public void setGPELInstanceID(URI gpelInstanceID) {
        this.gpelInstanceID = gpelInstanceID;
    }

    /**
     * Returns the name.
     * 
     * @return The name
     */
    public String getName() {
        // return this.name;
        return this.graph.getName();
    }

    /**
     * Sets name.
     * 
     * @param name
     *            The name to set.
     */
    public void setName(String name) {
        this.graph.setName(name);
    }

    /**
     * Returns the description.
     * 
     * @return The description
     */
    public String getDescription() {
        return this.graph.getDescription();
    }

    /**
     * Sets description.
     * 
     * @param description
     *            The description to set.
     */
    public void setDescription(String description) {
        this.graph.setDescription(description);
    }

    /**
     * @return The metadata, appinfo.
     */
    public XmlElement getMetadata() {
        return this.graph.getMetadata();
    }

    /**
     * @param metadata
     */
    public void setMetadata(XmlElement metadata) {
        this.graph.setMetadata(metadata);
    }

    /**
     * @return The output metadata, appinfo.
     */
    public XmlElement getInputMetadata() {
        return this.graph.getInputMetadata();
    }

    /**
     * @return The input metadata, appinfo.
     */
    public XmlElement getOutputMetadata() {
        return this.graph.getOutputMetadata();
    }

    /**
     * Returns the graph.
     * 
     * @return The graph
     */
    public WSGraph getGraph() {
        return this.graph;
    }

    /**
     * Sets graph.
     * 
     * @param graph
     *            The graph to set.
     */
    public void setGraph(WSGraph graph) {
        this.graph = graph;
    }

    /**
     * Returns the image.
     * 
     * @return The image
     */
    public BufferedImage getImage() {
        return this.image;
    }

    /**
     * Sets image.
     * 
     * @param image
     *            The image to set.
     */
    public void setImage(BufferedImage image) {
        this.image = image;
    }

    /**
     * Returns the gpelProcess.
     * 
     * @return The gpelProcess
     */
    public GpelProcess getGpelProcess() {
        return this.gpelProcess;
    }

    /**
     * Sets gpelProcess.
     * 
     * @param gpelProcess
     *            The gpelProcess to set.
     */
    public void setGpelProcess(GpelProcess gpelProcess) {
        this.gpelProcess = gpelProcess;
    }

    /**
     * Returns the workflowWSDL.
     * 
     * @return The workflowWSDL
     */
    public WsdlDefinitions getWorkflowWSDL() {
        return this.workflowWSDL;
    }

    /**
     * Sets workflowWSDL.
     * 
     * @param workflowWSDL
     *            The workflowWSDL to set.
     */
    public void setWorkflowWSDL(WsdlDefinitions workflowWSDL) {
        this.workflowWSDL = workflowWSDL;
    }

    /**
     * @return The set of WSDLs
     */
    public Map<String, WsdlDefinitions> getWSDLs() {

        Map<String, WsdlDefinitions> wsdls = new LinkedHashMap<String, WsdlDefinitions>();
        Map<WsdlDefinitions, String> ids = new HashMap<WsdlDefinitions, String>();
        // Use LinkedHashMap to preserve the order of WSDLs, which is useful for
        // some unit tests.

        for (WSNode node : GraphUtil.getNodes(this.graph, WSNode.class)) {
            WsdlDefinitions wsdl = node.getComponent().getWSDL();
            if (wsdls.containsValue(wsdl)) {
                String id = ids.get(wsdl);
                node.setWSDLID(id);
            } else {
                // Assign unique key
                String name = WSDLUtil.getWSDLName(wsdl);
                String id = StringUtil.convertToJavaIdentifier(name);
                while (wsdls.containsKey(id)) {
                    id = StringUtil.incrementName(id);
                }
                wsdls.put(id, wsdl);
                ids.put(wsdl, id);
                node.setWSDLID(id);
            }
        }
        return wsdls;
    }

    /**
     * This method is called by GPELClient during loading a workflow.
     * 
     * @param id
     * @param wsdl
     */
    public void addWSDL(String id, WsdlDefinitions wsdl) {
        logger.finest("id: " + id);
        this.wsdlMap.put(id, wsdl);
    }

    /**
     * Creates a node from a specified component and adds it to the graph.
     * 
     * @param component
     *            The specified component
     * @return The node added
     */
    public Node addNode(Component component) {
        Node node = component.createNode(this.graph);
        return node;
    }

    /**
     * Removes a specified node from the graph.
     * 
     * @param node
     *            The specified node
     * @throws GraphException
     */
    public void removeNode(Node node) throws GraphException {
        this.graph.removeNode(node);
    }

    /**
     * Imports a specified workflow to the current workflow.
     * 
     * @param workflow
     *            The specified workflow to import
     * @throws GraphException
     */
    public void importWorkflow(Workflow workflow) throws GraphException {
        this.graph.importGraph(workflow.getGraph());
    }

    /**
     * Returns the inputs of the workflow.
     * 
     * @return The inputs of the workflow.
     * @throws ComponentException
     */
    public List<WSComponentPort> getInputs() throws ComponentException {
        if (this.workflowWSDL == null) {
            throw new IllegalStateException();
        }
        WSComponent component = WSComponentFactory
                .createComponent(this.workflowWSDL);
        return component.getInputPorts();
    }

    /**
     * Returns the outputs of the workflow.
     * 
     * @return The outputs of the workflow.
     * @throws ComponentException
     */
    public List<WSComponentPort> getOutputs() throws ComponentException {
        if (this.workflowWSDL == null) {
            throw new IllegalStateException();
        }
        WSComponent component = WSComponentFactory
                .createComponent(this.workflowWSDL);
        return component.getOutputPorts();
    }

    /**
     * Returns the XML Text of the workflow.
     * 
     * @return The XML Text of the workflow
     */
    @Deprecated
    public String toXMLText() {
        return XMLUtil.xmlElementToString(toXML());
    }

    /**
     * Returns the XmlElement of the workflow.
     * 
     * @return The XmlElement of the workflow
     */
    public XmlElement toXML() {
        // This must be before graph.toXML() to set WSDL ID to each node.
        Map<String, WsdlDefinitions> wsdls = getWSDLs();

        XmlElement workflowElement = XMLUtil.BUILDER.newFragment(NS_XWF,
                WORKFLOW_TAG);

        // Version
        workflowElement.setAttributeValue(NS_XWF, VERSION_ATTRIBUTE,
                XBayaVersion.VERSION);

        // Date
        // TODO add modification time
        // XmlElement modifiedTimeElement = graphElement.addElement(
        // XgraphSchema.NS, "modifiedTime");
        // modifiedTimeElement.addChild(new GregorianCalendar().toString());

        // Graph
        workflowElement.addElement(this.graph.toXML());

        // WSDLs
        XmlElement wsdlsElement = workflowElement.addElement(NS_XWF, WSDLS_TAG);
        for (String id : wsdls.keySet()) {
            WsdlDefinitions wsdl = wsdls.get(id);
            XmlElement wsdlElement = wsdlsElement.addElement(NS_XWF, WSDL_TAG);
            wsdlElement.setAttributeValue(NS_XWF, ID_ATTRIBUTE, id);
            wsdlElement.setText(XMLUtil.xmlElementToString(wsdl.xml()));
        }

        // Image
        if (this.image != null) {
            try {
                ByteArrayOutputStream outStream = new ByteArrayOutputStream();
                ImageIO.write(this.image, XBayaConstants.PNG_FORMAT_NAME,
                        outStream);
                byte[] bytes = outStream.toByteArray();
                byte[] base64 = Base64.encodeBase64Chunked(bytes);

                XmlElement imageElement = workflowElement.addElement(NS_XWF,
                        IMAGE_TAG);
                imageElement.setText(new String(base64));
            } catch (IOException e) {
                // No image
                logger.caught(e);
            }
        }

        // BPEL
        if (this.gpelProcess != null) {
            XmlElement bpelElement = workflowElement.addElement(NS_XWF,
                    BPEL_TAG);
            bpelElement.setText(this.gpelProcess.xmlStringPretty());
        }

        // Workflow WSDL
        if (this.workflowWSDL != null) {
            XmlElement workflowWSDLElement = workflowElement.addElement(NS_XWF,
                    WORKFLOW_WSDL_TAG);
            workflowWSDLElement.setText(this.workflowWSDL.xmlStringPretty());
        }

        return workflowElement;
    }

    /**
     * Binds WSNodes to components
     * 
     * @throws ComponentException
     * @throws GraphException 
     */
    public void bindComponents() throws ComponentException, GraphException {
        // This map is to avoid creating multiple instances for the a component
        Map<WSComponentKey, WSComponent> components = new HashMap<WSComponentKey, WSComponent>();
        for (WSNode node : GraphUtil.getWSNodes(this.graph)) {
            String id = node.getWSDLID();
            logger.finest("id: " + id);
            WsdlDefinitions wsdl = this.wsdlMap.get(id);

            if (wsdl == null) {
                // XXX This happens while loading a workflow that is created by
                // the version 2.2.6_2 or below from GPEL.
                // Need to look for wsdl manually.
                // id should be look like
                // {http://www.extreme.indiana.edu/math/}Adder
                for (WsdlDefinitions w : this.wsdlMap.values()) {
                    QName name = WSDLUtil.getWSDLQName(w);
                    if (name.toString().equals(id)) {
                        wsdl = w;
                        break;
                    }
                }
            }

            QName portType = node.getPortTypeQName();
            if (portType == null) {
                // XXX This happens while parsing xwf created by the version
                // 2.2.6_1 or below.
                portType = WSDLUtil.getFirstPortTypeQName(wsdl);
            }
            String operation = node.getOperationName();
            if (operation == null) {
                // XXX This happens while parsing xwf created by the version
                // 2.2.6_1 or below.
                operation = WSDLUtil.getFirstOperationName(wsdl, portType);
            }
            WSComponentKey key = new WSComponentKey(id, portType, operation);

            WSComponent component;
            if (components.containsKey(key)) {
                component = components.get(key);
            } else {
                component = WSComponentFactory.createComponent(wsdl, portType,
                        operation);
                components.put(key, component);
            }
            node.setComponent(component);
        }
        
        this.graph.fixParameterNodes();
        GraphUtil.propagateTypes(this.graph);
    }

    /**
     * @see java.lang.Object#clone()
     */
    @Override
    public Workflow clone() {
        XmlElement originalXML = toXML();
        XmlElement newXML = XMLUtil.deepClone(originalXML);
        try {
            Workflow newWorkflow = new Workflow(newXML);
            return newWorkflow;
        } catch (GraphException e) {
            // This should not happen.
            throw new XBayaRuntimeException(e);
        } catch (ComponentException e) {
            // This should not happen.
            throw new XBayaRuntimeException(e);
        }
    }

    /**
     * @param graph
     * @return The workflow
     */
    public static Workflow graphToWorkflow(WSGraph graph) {
        Workflow workflow = new Workflow();
        workflow.setGraph(graph);

        workflow.setName(graph.getName());
        workflow.setDescription(graph.getDescription());
        return workflow;
    }

    /**
     * @param workflowElement
     * @throws GraphException
     * @throws ComponentException
     */
    private void parse(XmlElement workflowElement) throws GraphException,
            ComponentException {
        // Graph
        XmlElement graphElement = workflowElement
                .element(GraphSchema.GRAPH_TAG);
        this.graph = WSGraphFactory.createGraph(graphElement);

        XmlElement wsdlsElement = workflowElement.element(WSDLS_TAG);
        for (XmlElement wsdlElement : wsdlsElement.elements(null, WSDL_TAG)) {
            String wsdlText = wsdlElement.requiredText();
            WsdlDefinitions wsdl = WSDLUtil.stringToWSDL(wsdlText);
            String id = wsdlElement.attributeValue(NS_XWF, ID_ATTRIBUTE);
            if (id == null || id.length() == 0) {
                // xwf up to 2.2.6_2 doesn't have ID.
                id = WSDLUtil.getWSDLQName(wsdl).toString();
            }
            addWSDL(id, wsdl);
        }

        bindComponents();

        // Image
        XmlElement imageElement = workflowElement.element(IMAGE_TAG);
        if (imageElement != null) {
            String base64 = imageElement.requiredText();
            byte[] bytes = Base64.decodeBase64(base64.getBytes());
            try {
                this.image = ImageIO.read(new ByteArrayInputStream(bytes));
            } catch (IOException e) {
                // This should not happen and it's OK that image is broken. We
                // can reproduce it anytime.
                logger.caught(e);
            }
        }

        XmlElement bpelElement = workflowElement.element(BPEL_TAG);
        if (bpelElement != null) {
            try {
                String bpelString = bpelElement.requiredText();
                this.gpelProcess = new GpelProcess(XMLUtil
                        .stringToXmlElement(bpelString));
            } catch (RuntimeException e) {
                String error = "Failed to parse the BPEL document.";
                throw new GraphException(error, e);
            }
        }

        XmlElement workflowWSDLElement = workflowElement
                .element(WORKFLOW_WSDL_TAG);
        if (workflowWSDLElement != null) {
            try {
                String wsdlText = workflowWSDLElement.requiredText();
                this.workflowWSDL = new WsdlDefinitions(XMLUtil
                        .stringToXmlElement(wsdlText));
            } catch (RuntimeException e) {
                String error = "Failed to parse the workflow WSDL.";
                throw new GraphException(error, e);
            }
        }
    }
}

/*
 * Indiana University Extreme! Lab Software License, Version 1.2
 * 
 * Copyright (c) 2005-2007 The Trustees of Indiana University. All rights
 * reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1) All redistributions of source code must retain the above copyright notice,
 * the list of authors in the original source code, this list of conditions and
 * the disclaimer listed in this license;
 * 
 * 2) All redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the disclaimer listed in this license in
 * the documentation and/or other materials provided with the distribution;
 * 
 * 3) Any documentation included with all redistributions must include the
 * following acknowledgement:
 * 
 * "This product includes software developed by the Indiana University Extreme!
 * Lab. For further information please visit http://www.extreme.indiana.edu/"
 * 
 * Alternatively, this acknowledgment may appear in the software itself, and
 * wherever such third-party acknowledgments normally appear.
 * 
 * 4) The name "Indiana University" or "Indiana University Extreme! Lab" shall
 * not be used to endorse or promote products derived from this software without
 * prior written permission from Indiana University. For written permission,
 * please contact http://www.extreme.indiana.edu/.
 * 
 * 5) Products derived from this software may not use "Indiana University" name
 * nor may "Indiana University" appear in their name, without prior written
 * permission of the Indiana University.
 * 
 * Indiana University provides no reassurances that the source code provided
 * does not infringe the patent or any other intellectual property rights of any
 * other entity. Indiana University disclaims any liability to any recipient for
 * claims brought by any other entity based on infringement of intellectual
 * property rights or otherwise.
 * 
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS.
 * INDIANA UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS",
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION
 * GENERATED USING SOFTWARE.
 */

